#!/usr/bin/perl -w
#------------------------------------------------------------------------
# mmc.pl
#
# Demo of decoding the memory size from the bit array from
# a CSD register of an MMC card.
#
# This follows the algorithm on page 3-15 of the PDF document
# "MultiMediaCard Product Manual" rev 5.1, SanDisk corp.
# Perl uses the same mnemonics like "MULT", ... as SanDisk,
# to help follow this convoluted decoding process.
#
# card specific device register: 16 bytes, CRC and trailing FFs:
# CSD: 48,0E,01,2A,0F,F9,81,EA,EC,B1,01,E1,8A,40,00,BB,82,9C,FF,FF
#
# Results of running this script for 16M card: 16089088 = 0xF58000 bytes.
#
# This agrees with experimental reads until an out-of-bounds error
# is returned - success in accessing from 00000000 to 00F57E00, then
# the read returns an error, as expected.
#
#------------------------------------------------------------------------

use strict;

my(@csd, $memcap);
my($MULT, $BLOCK_LEN, $READ_BL_LEN, $BLOCK_NR, $C_SIZE, $C_SIZE_MULT);

  @csd = (72,14,01,42,15,249,129,234,236,177,01,225,138,64,00,187);

  $C_SIZE_MULT = &bits(49, 47);
  $MULT = $C_SIZE_MULT << 3;

  $READ_BL_LEN = &bits(83, 80);
  $BLOCK_LEN = 1 << $READ_BL_LEN;

  $C_SIZE = &bits(73, 62);
  $BLOCK_NR = ($C_SIZE+1) * $MULT;

  $memcap = $BLOCK_NR * $BLOCK_LEN;

  print "Mem cap: $memcap\n";

#------- end of main ------------------------------
#
# Return the number consisting of the specified bits

sub bits{
  my($i, $start, $end, $result, $shift, $bit);

  $start = $_[0];
  $end = $_[1];

  $result = 0;
  $i = $start;
  $shift = $start - $end;

  while($i >= $end){
    $bit = &bit($i);
    $result = $result + ($bit << $shift);
    $i--;
    $shift--;
  }
  $result;
}

#--------------------------------------------------
#
# Return the bit (0 or 1) at the argument location from the CSD array

sub bit{
  my($bitofbyte, $bit, $byte, $bytenumber, $bitnumber);

  $bitnumber = $_[0];
  $bytenumber = 15 - (int($bitnumber/8));

  $byte = $csd[$bytenumber];
  $bitofbyte = 7 - ((127 - $bitnumber) - (8 * $bytenumber));
  $bit = ($byte >> $bitofbyte) & 1;
  if ($bit > 0){ $bit = 1; }

  $bit;
}

#--------------------------------------------------
